/**
 *	Height Map Terrain - Generator and Viewer
 *	@author Everett Morse
 *	
 *	This is made for Brigham Young University's CS 658 course "Computer Generated 
 *	Natural Phenomenon" (graduate level).
 *	
 *	Graded Features:
 *	- Implement square-diamond terrain generation and display (85%)
 *	- Implement Muskgrave's Thermal Weathering algorithm (5%)
 *	- Rotatable display (2%)
 *	- Coloration based on altitude (2%)
 *	- "Cool" GUI features (5-10%)
 *		- Change algorithm parameters
 *		- Stepper for Square-Diamond and Thermal Weathering
 *		- Water level display
 *	
 *	Other features that would count for grading but that I have not done:
 *	- Implement Muskgrave's Hydraulic Weathering algorithm (10%)
 *	- Coloring based on face direction (2%)
 *	- Coloring based on face angle (2%)
 *	- Vary the "roughness" based on altitude (5%)
 *	
 *	Other ungraded features:
 *	- Can load/save a height map
 *	
 *	TODO:
 *	- Do error checking in the load/save code to print a message vs. segfault on 
 *	  null file pointer.
 *	- Sanitize the filename input from GLUi, which apparently allows things like tab
 *	  (I'm guessing it was a tab, it was some control character.)
 *	- Maybe have a nice popup to select file location
 *	- Find the OpenGL command to do culling on the pixel vs. polygon level.  On
 *	  my MacBook this happens automatically, on my desktop it doesn't.
 *	- Have a nicer way to notify the user, for file errors and for weathering complete.
 *	- Hydraulic weathering
 *	- Roughness vary by altitude (lacunarity)
 * 
 *	Organization:
 *	This file contains the driver that handles the GUI and main OpenGL display 
 *	functions and launches everything.  Each algorithm is in it's own file, as
 *	well as separate files to render a height map and the water level box.
 *
 *	Display:
 *	The origin is the center of the terrain.  Terrain will be 1000x1000, with an
 *	up and down range of 500.  The camera, therefore, should be set above and at
 *	some angle, so should the lights.
 *
 *	Note: Some of this code is copied from a moon phase viewer someone else wrote,
 *	but that I modified to add in a GLUi-based interface.
 */

/* Standard Includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define PI 3.14159265

/* Open GL, GLUT, GLUi */
//#include <GLUT/glut.h>	//Mac - GLUi will include this
//#include <OpenGL/glu.h>	//Mac - not needed
//#include <OpenGL/gl.h>	//Mac - not needed
//#include <GL\glut.h>		//Windows
#include "GL/glui.h"

/* Project Includes */
#include "MidPoint.h"
#include "ThermalWeather.h"
#include "texture.h"


/* Settings */
#include "Constants.h"
#define FOVY 65.0
double zFar = 1500;
double zNear = 1;
//Copied from the Moon Phase program
float LightNoAmbient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
float LightPartialAmbient[] = { 0.1f, 0.1f, 0.1f, 1.0f };
float LightDiffuse[] = { 0.8f, 0.8f, 0.8f, 1.0f };
float LightSpecular[] ={ 0.5f, 0.5f, 0.5f, 1.0f };
float LightPosition[] = { -1.5f, 1.0f, 4.0f, 1.0f };
float LightNoEmission[] = {0.0f, 0.0f, 0.0f, 0.5f};
float LightFullEmission[] = {1.0f, 1.0f, 1.0f, 1.0f};
float LightSunEmission[] = {1.0f, 1.0f, 0.0f, 1.0f};


/* Terrain */
HeightMap *hmap;
int colorScheme = 0;
float rotation[16] = {	1.0, 0.0, 0.0, 0.0,
						0.0, 1.0, 0.0, 0.0, 
						0.0, 0.0, 1.0, 0.0,
						0.0, 0.0, 0.0, 1.0 };
int mapSize = 129;
float mapRoughness = 0.5f;
int water = 0;
float waterLevel = 0;
GLuint WATER_TEXTURE;
int renderMode = MESH;

/* Weathering */
ThermalState* thermal = NULL;
int thermalAngle = 45;
float thermalMove = 0.2f;

/* GLUI interface */
int windowId;			//Main Window Id
GLUI *glui;
GLUI_Rotation *rotator;
int smoothShade = 1;
int cull = 1;
char filename[sizeof(GLUI_String)];
//Button Ids
#define B_QUIT 1
#define B_ROTATION 2
#define B_RESET_ROTATION 3
#define B_COLOR 4
#define B_SHADING 5
#define B_RESET_MAP 6
#define B_STEP_MAP 7
#define B_RUN_MAP 8
#define B_THERMAL_VAR 9
#define B_STEP_THERMAL 10
#define B_RUN_THERMAL 11
#define B_WATER 12
#define B_WATER_LEVEL 13
#define B_CULL 14
#define B_LOAD 15
#define B_SAVE 16
#define B_SHOW_CELLS 17


/**
 * Set up the camera position and perspective. (Called by draw scene.)
 */
void ResetPerspective(){
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();									
	gluPerspective(FOVY, (float)WIDTH/(float)HEIGHT, zNear, zFar);
	
	glMatrixMode(GL_MODELVIEW);							
	glLoadIdentity();	
	//gluLookAt(700, 0, 700,		//Eye -- looking down X axis from some elevation
	//		  0, 0, 0,			//Center -- center of terrain
	//		  0, 0, 1);			//Up -- +Z is up
	gluLookAt(0, -700, 700,		//Eye -- looking down Z axis from some elevation
			  0, 0, 0,			//Center -- center of terrain
			  0, -1, 0);		//Up -- -Y is up

	//glViewport(0, 0, WIDTH, HEIGHT);
	GLUI_Master.auto_set_viewport();
}

/**
 * Set up lighting. (Called by draw scene.)
 */
void InitLighting(void){
	
	// put light at a 45 degree angle up, right between X and Y axis
	//float light_pos[] = { 700.0f, 700.0f, 700.0f, 1.0f };
	float light_pos[] = { 1000.0f, -1000.0f, 1000.0f, 1.0f };
	
	glLightfv(GL_LIGHT0, GL_AMBIENT, LightPartialAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, LightDiffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, LightSpecular);
	
	glMaterialfv(GL_FRONT, GL_SPECULAR, LightSpecular);
	glMateriali(GL_FRONT, GL_SHININESS, 128);
	
	glLightfv( GL_LIGHT0, GL_POSITION, light_pos );
	
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
}

/**
 * Set up OpenGL (called by main)
 */
void InitGL(void) {
	ResetPerspective();
	
	if( smoothShade )
		glShadeModel(GL_SMOOTH);
	else
		glShadeModel(GL_FLAT);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	if( cull ) {
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
		glDepthMask(GL_TRUE);
	}
	glCullFace(GL_BACK);
	
	InitLighting();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	//glDepthFunc(GL_LEQUAL);								
	//glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	
	
	/*
	if(wireframe)
		glPolygonMode(GL_FRONT,GL_LINE);
	else
		glPolygonMode(GL_FRONT,GL_FILL);
	//*/
}

/**
 * Called by GLUT to draw the window contents (other than GLUi sub-window)
 */
void DrawScene(void) {
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	
	ResetPerspective();
	InitLighting();
	
	//Rotate, then draw the height map terrain
	glPushMatrix();
	glMultMatrixf(rotation);
	
	
	glRotatef(90.0f, 1.0f, 0.0f, 0.0f);	//rotate for terrain to be XY (+ towards viewer and off right)
	//glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
	//glRotatef(90.0f, 0.0f, 0.0f, 1.0f);
	
	hmap_draw(hmap);
	if( water ) {
		glColor3f(0.3f, 0.3f, 1.0f);
		glBegin(GL_QUADS);
		
		//N - br, tr, tl, bl from the perspective of facing the north plane
		glVertex3f(-500.0f, -500.0f, hmap->min);
		glVertex3f(-500.0f, -500.0f, waterLevel);
		glVertex3f(-500.0f, 500.0f, waterLevel);
		glVertex3f(-500.0f, 500.0f, hmap->min);
		//E
		glVertex3f(-500.0f, 500.0f, hmap->min);
		glVertex3f(-500.0f, 500.0f, waterLevel);
		glVertex3f(500.0f, 500.0f, waterLevel);
		glVertex3f(500.0f, 500.0f, hmap->min);
		//S
		glVertex3f(500.0f, 500.0f, hmap->min);
		glVertex3f(500.0f, 500.0f, waterLevel);
		glVertex3f(500.0f, -500.0f, waterLevel);
		glVertex3f(500.0f, -500.0f, hmap->min);
		//W
		glVertex3f(500.0f, -500.0f, hmap->min);
		glVertex3f(500.0f, -500.0f, waterLevel);
		glVertex3f(-500.0f, -500.0f, waterLevel);
		glVertex3f(-500.0f, -500.0f, hmap->min);
		glEnd();
		
		//Top
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, WATER_TEXTURE);
		glBegin(GL_QUADS);
		glColor3f(1.0f, 1.0f, 1.0f);
		
		glTexCoord2f(1.0, 0.0);
		glVertex3f(500.0f, 500.0f, waterLevel);
		glTexCoord2f(1.0, 1.0);
		glVertex3f(-500.0f, 500.0f, waterLevel);
		glTexCoord2f(0.0, 1.0);
		glVertex3f(-500.0f, -500.0f, waterLevel);
		glTexCoord2f(0.0, 0.0);
		glVertex3f(500.0f, -500.0f, waterLevel);
		
		glEnd();
		glDisable(GL_TEXTURE_2D);
	}
	
	glPopMatrix();
		
	glutSwapBuffers();
}

/**
 * Call when the terrain or view angle has changed
 */
void Repaint() {
	glutSetWindow(windowId);
	glutPostRedisplay();
}


void setThermalVars() {
	double angle, run;
	
	if( thermal == NULL )
		return;
	
	angle = thermalAngle * PI / 180;		//convert to radians
	run = TERRAIN_WIDTH / (thermal->map->size - 1);	//distance between points
	
	/*
		rise / run = tan(angle)
		T is the value of rise, which is max in order to meet angle of repose
		rise = tan(angle) * run
	*/
	
	thermal->T = (int)( tan(angle) * run );
	thermal->C = thermalMove;
	
	//printf("T=%d\n", thermal->T);
}

void loadFile() {
	FILE *file;
	int i,j, size;
	char buf[16];
	
	file = fopen(filename,"r");
	fgets(buf, 16, file);
	size = atoi(buf);
	
	//Initialize and reset height map
	if( hmap->size != size ) {
		mapSize = size;
		
		hmap_free(hmap);
		if( thermal != NULL)
			thermal_free(thermal);
		thermal = NULL;
		
		hmap = hmap_create(mapSize);
	} else {
		hmap_init(hmap);
	}
	//say we're done already
	hmap->stepLength = 1;
	hmap->square = 0;
	
	
	printf("Reading file with size=%d\n", size);
	
	for(j = 0; j < size; j++) {
		for(i = 0; i < size; i++) {
			fread((void*)buf, sizeof(char), 16, file);	//15 chars number, 1 char space
			hmap->heights[i][j] = (float)atof(buf);
			//printf("%5.1f ", hmap->heights[i][j]);
		}
		buf[0] = getc(file);
		if( buf[0] != '\n' ) {
			printf("Error! Expected newline");
			return;
		}
	}
	
	fclose(file);
	printf("Done\n");
}

void saveFile() {
	FILE *file;
	int i,j;
	
	file = fopen(filename,"w+");
	fprintf(file,"%d\n",hmap->size);
	for(j = 0; j < hmap->size; j++) {
		for(i = 0; i < hmap->size; i++) {
			fprintf(file,"%15.10f ",hmap->heights[i][j]);
		}
		fprintf(file,"\n");
	}
	
	fclose(file);
}

/**
 * Called by GLUi when a button has been pressed.
 */
void ControlButton(int id) {
	int c;
	
	switch(id) {
		case B_QUIT:
			exit(0);
			break;
		case B_ROTATION:
			Repaint();
			break;
		case B_RESET_ROTATION:
			rotator->reset();
			Repaint();
			break;
		case B_COLOR:
		case B_SHOW_CELLS:
			hmap_decompile(hmap);	//force it to regenerate the display list
			Repaint();
			break;
		case B_SHADING:
			if( smoothShade )
				glShadeModel(GL_SMOOTH);
			else
				glShadeModel(GL_FLAT);
			Repaint();
			break;
		case B_CULL:
			if( cull ) {
				glEnable(GL_DEPTH_TEST);
				glEnable(GL_CULL_FACE);
				glDepthMask(GL_TRUE);
			} else {
				glDisable(GL_DEPTH_TEST);
				glDisable(GL_CULL_FACE);
			}
			Repaint();
			break;
		case B_RESET_MAP:
			//Update all the config settings, then start fresh
			if( mapSize != hmap->size ) {
				hmap_free(hmap);
				if( thermal != NULL)
					thermal_free(thermal);
				thermal = NULL;
				
				hmap = hmap_create(mapSize);	//includes call to init
			} else {
				hmap_init(hmap);	//resets heightmap
				hmap_decompile(hmap);	//force it to regenerate the display list
			}
			hmap->roughness = mapRoughness;
			
			Repaint();
			break;
		case B_STEP_MAP:
			hmap_step(hmap);
			hmap_decompile(hmap);	//force it to regenerate the display list
			Repaint();
			break;
		case B_RUN_MAP:
			while( !hmap_step(hmap) );
			hmap_decompile(hmap);	//force it to regenerate the display list
			Repaint();
			break;
		case B_THERMAL_VAR:
			setThermalVars();
			break;
		case B_STEP_THERMAL:
			if( thermal == NULL ) {
				thermal = thermal_create(hmap);
				setThermalVars();
			}
			thermal_step(thermal);
			
			hmap_decompile(hmap);	//force it to regenerate the display list
			Repaint();
			break;
		case B_RUN_THERMAL:
			if( thermal == NULL ) {
				thermal = thermal_create(hmap);
				setThermalVars();
			}
			
			//Only do 20 steps at a time (to avoid infinite loops)
			for(c = 0; c < 20; c++) {
				if( thermal_step(thermal) ) {
					printf("Thermal Weathering settled\n", c);
					break;	//reported done, so stop
				}
			}
			
			hmap_decompile(hmap);	//force it to regenerate the display list
			Repaint();
			break;
		case B_WATER:
			Repaint();
			break;
		case B_WATER_LEVEL:
			if( hmap != NULL && waterLevel < hmap->min ) {
				waterLevel = hmap->min;
				glui->sync_live();
			}
			Repaint();
			break;
		case B_LOAD:
			loadFile();
			glui->sync_live();
			Repaint();
			break;
		case B_SAVE:
			saveFile();
			break;
		default:
			break;
	}
}

void InitGUI(int windowId) {
	GLUI_Listbox *box;
	GLUI_Spinner *spin;
	
	glui = GLUI_Master.create_glui_subwindow(windowId, GLUI_SUBWINDOW_LEFT);
	
	GLUI_Rollout *p1 = glui->add_rollout("Generate", true);
	GLUI_Rollout *p2 = glui->add_rollout("Thermal Weathering", false);
	GLUI_Rollout *p5 = glui->add_rollout("Hydraulic Weathering", false);
	GLUI_Rollout *p3 = glui->add_rollout("Display", true);
	GLUI_Rollout *p4 = glui->add_rollout("File System", true);
	glui->add_button("Quit", B_QUIT, ControlButton);
	
	//Generate
	box = glui->add_listbox_to_panel(p1, "Size", (int*)&mapSize);
	box->add_item(5, "5x5");
	box->add_item(9, "9x9");
	box->add_item(17, "17x17");
	box->add_item(33, "33x33");
	box->add_item(65, "65x65");
	box->add_item(129, "129x129");
	box->add_item(257, "257x257");
	box->add_item(513, "513x513");
	box->add_item(1025, "1025x1025");
	box->add_item(2049, "2049x2049");
	
	spin = glui->add_spinner_to_panel(p1, "Roughness", GLUI_SPINNER_FLOAT, &mapRoughness);
	spin->set_float_limits(0.001, 1.0);
	
	glui->add_button_to_panel(p1, "Reset", B_RESET_MAP, ControlButton);
	glui->add_button_to_panel(p1, "Step", B_STEP_MAP, ControlButton);
	glui->add_button_to_panel(p1, "Run", B_RUN_MAP, ControlButton);
	
	//Thermal Weathering
	box = glui->add_listbox_to_panel(p2, "Angle of Repose", (int*)&thermalAngle, B_THERMAL_VAR, ControlButton);
	box->add_item(10, "10 deg");
	box->add_item(15, "15 deg");
	box->add_item(20, "20 deg");
	box->add_item(30, "30 deg");
	box->add_item(45, "45 deg");
	box->add_item(60, "60 deg");
	
	spin = glui->add_spinner_to_panel(p2, "Move Percent", GLUI_SPINNER_FLOAT, &thermalMove, B_THERMAL_VAR, ControlButton);
	spin->set_float_limits(0.001, 1.0);
	
	glui->add_button_to_panel(p2, "Step", B_STEP_THERMAL, ControlButton);
	glui->add_button_to_panel(p2, "Step+20", B_RUN_THERMAL, ControlButton);
	
	//Display
	rotator = glui->add_rotation_to_panel(p3, "Rotation", rotation, B_ROTATION, ControlButton);
	glui->add_button_to_panel(p3, "Reset", B_RESET_ROTATION, ControlButton);
	
	box = glui->add_listbox_to_panel(p3, "Mode", (int*)&renderMode, B_SHOW_CELLS, ControlButton);
	box->add_item(MESH,	"Peaks Mesh");
	box->add_item(SINGLE_CELL, "Ind. Cells");
	box->add_item(SINGLE_CELL2, "Ind. Cells 2");
	box->add_item(LARGE_CELL, "Lrg Cells");
	box->add_item(LARGE_CELL2, "Lrg Cells 2");
	
	box = glui->add_listbox_to_panel(p3, "Color", (int*)&colorScheme, B_COLOR, ControlButton);
	box->add_item(-1, "Solid");
	box->add_item(0, "Wooded Peaks");
	box->add_item(1, "Greens");
	box->add_item(2, "Blues");
	box->add_item(3, "Sand");
	box->add_item(4, "Rocks");
	
	glui->add_checkbox_to_panel(p3, "Smooth Shading", (int*)&smoothShade, B_SHADING, ControlButton);
	glui->add_checkbox_to_panel(p3, "Culling", (int*)&cull, B_CULL, ControlButton);
	
	glui->add_checkbox_to_panel(p3, "Water", (int*)&water, B_WATER, ControlButton);
	spin = glui->add_spinner_to_panel(p3, "Water Level", GLUI_SPINNER_FLOAT, &waterLevel, B_WATER_LEVEL, ControlButton);
	spin->set_float_limits(-500.f, 500.0f);
	
	//File System
	GLUI_EditText *txt = glui->add_edittext_to_panel(p4, "File Name:", GLUI_EDITTEXT_TEXT, filename);
	glui->add_button_to_panel(p4, "Load", B_LOAD, ControlButton);
	glui->add_button_to_panel(p4, "Save", B_SAVE, ControlButton);
	
	//Set default
	txt->set_text("map.txt");
	
	glui->set_main_gfx_window(windowId);
	glui->sync_live();
}

int main(int argc, char** argv){
	
	glutInit(&argc, argv); 
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH); 
	glutInitWindowSize(WIDTH, HEIGHT); 
	glutInitWindowPosition(200, 100); 
	windowId = glutCreateWindow("Terrain - Generater & Viewer");
	InitGUI(windowId);
	
	InitGL(); 
	
	//Texture loading
	loadTextures();
	
	glutDisplayFunc(DrawScene); 
	GLUI_Master.set_glutIdleFunc(NULL);
	//GLUI_Master.set_glutKeyboardFunc(KeyDown);
	//GLUI_Master.set_glutSpecialFunc(SpecialKeyDown);
	//GLUI_Master.set_glutMouseFunc( myGlutMouse );
	//GLUI_Master.set_glutReshapeFunc( myGlutReshape );
	
	//Create a heightmap to display
	hmap = hmap_create(mapSize);
	hmap->roughness = mapRoughness;
	//hmap_print(hmap);
	
	
	Repaint();
	glutMainLoop();
	
	//thermal_free(thermal);
	hmap_free(hmap);
	return 0;
}