#include "HeightMap.h"
#include "Constants.h"
#include <OpenGL/gl.h>
#include <stdio.h>


//Static function prototype
void hmap_render(HeightMap*);
void hmap_renderCells(HeightMap*);
void hmap_renderCells2(HeightMap*);
void hmap_heightColor(float height, float min, float max);

//Variables
extern int colorScheme;		//set in Driver.c by the gui
extern int renderMode;		//whether to show a column per cell or make a terain mesh



/* Color Schemes */
//Humbug.  I can't seem to get a good array of arrays or array of color scheme structs that I can
//initialize here.  No one posts help that obviously applies.  So I'm going to do something ghetto.

//Wooded Peaks
int numColors0 = 6;
float colors0[] = {
	0.1f, 0.1f, 0.05f,		//bottom is dark
	0.0f, 0.5f, 0.0f,		//shades of green
	0.3f, 1.0f, 0.3f,		//light green
	0.6f, 0.6f, 0.4f,		//bare rock brown
	0.9f, 0.9f, 0.9f,		//snow white
	1.0f, 1.0f, 1.0f		//pure white at top
};
float cuttoffs0[] = {		//these are percentages, scaled to max and min height of the given map
	0.03f,
	0.5f,
	0.7f,
	0.75f,
	0.98f,
	1.0f
};
//Greens
int numColors1 = 2;
float colors1[] = {
	0.1f, 0.3f, 0.1f,
	0.2f, 1.0f, 0.2f
};
float cuttoffs1[] = {
	0.03f,
	1.0f
};
//Blues
int numColors2 = 2;
float colors2[] = {
	0.2f, 0.2f, 0.5f,
	0.7f, 0.7f, 1.0f
};
float cuttoffs2[] = {
	0.03f,
	1.0f
};
//Sand
int numColors3 = 2;
float colors3[] = {
	0.2f, 0.2f, 0.1f,
	0.7f, 0.7f, 0.5f
};
float cuttoffs3[] = {
	0.03f,
	1.0f
};
//Rocks
int numColors4 = 7;
float colors4[] = {
	0.1f, 0.1f, 0.05f,		//bottom is dark
	0.5f, 0.5f, 0.3f,		//brown
	0.4f, 0.4f, 0.4f,		//gray
	0.0f, 0.0f, 0.0f,		//black
	0.6f, 0.6f, 0.4f,		//bare rock brown
	0.85f, 0.85f, 0.85f,	//white
	0.4f, 0.4f, 0.3f		//some brown
};
float cuttoffs4[] = {
	0.03f,
	0.3f,
	0.6f,
	0.7f,
	0.8f,
	0.93f,
	1.0f
};

//These will point to whichever is actually in use
int scheme = -1;
int numColors;
float *colors;
float *cuttoffs;



/**
 * Displays the height map as a terrain mesh
 */
void hmap_draw(HeightMap* map) {
	//Make sure to choose the right color scheme
	if( scheme != colorScheme && colorScheme != -1 ) {
		switch(colorScheme) {
			case 0:		numColors = numColors0; colors = colors0; cuttoffs = cuttoffs0; break;
			case 1:		numColors = numColors1; colors = colors1; cuttoffs = cuttoffs1; break;
			case 2:		numColors = numColors2; colors = colors2; cuttoffs = cuttoffs2; break;
			case 3:		numColors = numColors3; colors = colors3; cuttoffs = cuttoffs3; break;
			case 4:		numColors = numColors4; colors = colors4; cuttoffs = cuttoffs4; break;
		}
		scheme = colorScheme;
	}
	
	
	//Make sure we have a display list
	if( map->displayId == 0 )
		hmap_compile(map);
	
	//Now draw it
	glCallList(map->displayId);
}

/**
 * Creates the display list
 */
void hmap_compile(HeightMap* map) {
	if( map->displayId != 0 )
		hmap_decompile(map);	//properly free old list
	
	map->displayId = DL_HMAP;	//well, just keep using the same thing I guess...
	glNewList(map->displayId, GL_COMPILE);
	
	switch( renderMode ) {
		case MESH:
			hmap_render(map);
			break;
		case SINGLE_CELL:
		case SINGLE_CELL2:
			hmap_renderCells(map);
			break;
		case LARGE_CELL:
		case LARGE_CELL2:
			hmap_renderCells2(map);
			break;
		default:
			printf("Invalid render mode!");
			break;
	}
	
	glEndList();
	
	//Could use... ?
	//glNewList(glGenLists(1), GL_COMPILE_AND_EXECUTE);
}

/**
 * Properly frees the display list
 */
void hmap_decompile(HeightMap* map) {
	if( map->displayId != 0 ) {
		//TODO:  free existing display list
		map->displayId = 0;
	}
}

/**
 * Draws the map.  This is used when compiling the display list.
 * 
 * This makes a mesh where each row is a trianlge strip.  A row
 * in the mesh is the series of triangles between two rows of
 * points in the height map.
 * 
 * Strips are build like this:
 *		1-2
 *		|/|
 *		3-4
 *		|/|
 *		5-6
 */
void hmap_render(HeightMap* map) {
	float dxy;
	int i, j, inc;
	float min, max;	//min and max heights needed for color calculations
	
	//Change in x and y for each cell
	dxy = TERRAIN_WIDTH / (map->size-1);
	
	//Set Color
	if( colorScheme == -1 ) {
		glColor3f(0.7, 0.7, 0.5);
	} else {
		//Need to find min and max
		min = max = 0;
		for(i = 0; i < map->size; i++) {
			for(j = 0; j < map->size; j++) {
				float h = map->heights[i][j];
				if( h < min )
					min = h;
				if( h > max )
					max = h;
			}
		}
		map->min = min;	//cache this so the water box renderer can use it
	}
	
	inc = map->stepLength;	//finished terrain -> 1, otherwise skip unset heights
	if( inc != 1 && map->square == 0 )
		inc /= 2;	//have done the diamond step, so show results of that
	
	for(i = inc; i < map->size; i += inc) {
		//Starts after the first row and draw space between i-1 and i
		//(note: we're really do columns if you think of a normal XY plane)
		
		glBegin(GL_TRIANGLE_STRIP);
		for(j = 0; j < map->size; j += inc) {
			if( colorScheme != -1 )
				hmap_heightColor(map->heights[i-inc][j], min, max);
			glVertex3f((i-inc)*dxy-(TERRAIN_WIDTH/2), j*dxy-(TERRAIN_WIDTH/2), map->heights[i-inc][j]);
			
			if( colorScheme != -1 )
				hmap_heightColor(map->heights[i][j], min, max);
			glVertex3f(i*dxy-(TERRAIN_WIDTH/2), j*dxy-(TERRAIN_WIDTH/2), map->heights[i][j]);
		}
		glEnd();
	}
}

void hmap_heightColor(float height, float min, float max) {
	float percent, range;
	int upperIndex;
	
	percent = (height - min) / (max - min);
	for(upperIndex = 0; upperIndex < numColors; upperIndex++) {
		if( percent < cuttoffs[upperIndex])
			break;	//found closest upper bound
	}
	
	if( upperIndex == numColors ) {
		//Is the max, so use highest color
		glColor3f(colors[(upperIndex-1)*3 + 0], colors[(upperIndex-1)*3 + 1], colors[(upperIndex-1)*3 + 2]);
		return;
	}
	
	if( upperIndex == 0 ) {
		//Below first bound, so just use lowest color
		glColor3f(colors[0], colors[1], colors[2]);
		return;
	}
	
	//Otherwise we can interpolate between lower and upper bound colors
	percent -= cuttoffs[upperIndex - 1];
	range = cuttoffs[upperIndex] - cuttoffs[upperIndex - 1];
	percent = percent / range;
	
	//range = upper - lower;
	//value = lower + percent * range
	float r, g, b;
	r = colors[(upperIndex - 1)*3 + 0] + percent * (colors[(upperIndex)*3 + 0] - colors[(upperIndex - 1)*3 + 0]);
	g = colors[(upperIndex - 1)*3 + 1] + percent * (colors[(upperIndex)*3 + 1] - colors[(upperIndex - 1)*3 + 1]);
	b = colors[(upperIndex - 1)*3 + 2] + percent * (colors[(upperIndex)*3 + 2] - colors[(upperIndex - 1)*3 + 2]);
	
	glColor3f(r, g, b);
}

/**
 * This shows each cell as a box column at the correct altitude.  This just makes
 * visualization easier.  Colors are interpolated same as with the mesh.
 */
void hmap_renderCells(HeightMap* map) {
	float dxy;
	int i, j;
	float min, max;	//min and max heights needed for color calculations
	
	//Change in x and y for each cell
	dxy = TERRAIN_WIDTH / (map->size);
	
	//Set Color
	if( colorScheme == -1 ) {
		glColor3f(0.7, 0.7, 0.5);
	} else {
		//Need to find min and max
		min = max = 0;
		for(i = 0; i < map->size; i++) {
			for(j = 0; j < map->size; j++) {
				float h = map->heights[i][j];
				if( h < min )
					min = h;
				if( h > max )
					max = h;
			}
		}
		map->min = min;	//cache this so the water box renderer can use it
	}
	
	glBegin(GL_QUADS);
	for(i = 0; i < map->size; i++) {
		for(j = 0; j < map->size; j++) {
			float altitude = map->heights[i][j];
			
			if( renderMode == SINGLE_CELL2 && map->square == 0 && altitude == 0 )
				continue;	//skip this cell on the diamond step if it is zero, b/c it's probably unset
			
			if( colorScheme != -1 )
				hmap_heightColor(altitude, min, max);
			
			//N - br, tr, tl, bl from the perspective of facing the north plane
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			//E
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			//S
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			//W
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			
			//Top
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
		}
	}
	glEnd();
}

/**
 * This shows each cell as a box column at the correct altitude.  This just makes
 * visualization easier.  Colors are interpolated same as with the mesh.
 * Modification: use the stepLength so we don't have tiny cells for earlier steps.
 */
void hmap_renderCells2(HeightMap* map) {
	float dxy;
	int i, j, inc;
	float min, max;	//min and max heights needed for color calculations
	
	//Set Color
	if( colorScheme == -1 ) {
		glColor3f(0.7, 0.7, 0.5);
	} else {
		//Need to find min and max
		min = max = 0;
		for(i = 0; i < map->size; i++) {
			for(j = 0; j < map->size; j++) {
				float h = map->heights[i][j];
				if( h < min )
					min = h;
				if( h > max )
					max = h;
			}
		}
		map->min = min;	//cache this so the water box renderer can use it
	}
	
	inc = map->stepLength;	//finished terrain -> 1, otherwise skip unset heights
	if( inc != 1 && map->square == 0 )
		inc /= 2;	//have done the diamond step, so show results of that
	
	//Change in x and y for each cell
	int steps = ((map->size-1)/inc + 1);
	dxy = TERRAIN_WIDTH / steps;
	
	
	glBegin(GL_QUADS);
	for(i = 0; i < steps; i++) {
		for(j = 0; j < steps; j++) {
			float altitude = map->heights[i*inc][j*inc];
			
			if( renderMode == LARGE_CELL2 && map->square == 0 && altitude == 0 )
				continue;	//skip this cell on the diamond step if it is zero, b/c it's probably unset
			
			if( colorScheme != -1 )
				hmap_heightColor(altitude, min, max);
			
			//N - br, tr, tl, bl from the perspective of facing the north plane
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			//E
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			//S
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			//W
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), min);
			
			//Top
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), (j+1)*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f(i*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
			glVertex3f((i+1)*dxy - (TERRAIN_WIDTH/2), j*dxy - (TERRAIN_WIDTH/2), altitude);
		}
	}
	glEnd();
}