#include <stdlib.h>
#include <stdio.h>
#include "MidPoint.h"
#include "Util.h"
#include "Constants.h"



/**
 * Allocates the space and calls init.
 * @precondition  size = 2^N + 1 for some N
 */
HeightMap* hmap_create(int size) {
	HeightMap *map;
	
	map = (HeightMap*)malloc(sizeof(HeightMap));
	map->size = size;
	map->heights = (float**)malloc_a2d(size, size, sizeof(float));
	map->displayId = 0;
	map->roughness = 0.5;	//range is halved at each step
	map->range = TERRAIN_HEIGHT;
	map->min = 0;
	
	hmap_init(map);
	return map;
}



/**
 * Will reset a map or init for the first time
 */
void hmap_init(HeightMap* map) {
	int i, j;
	
	//If we've compiled the display list, free it
	if( map->displayId != 0 ) {
		hmap_decompile(map);
	}
	
	//Init to zero
	for(i = 0; i < map->size; i++) {
		for(j = 0; j < map->size; j++) {
			map->heights[i][j] = 0;
		}
	}
	
	//reset to first step
	map->stepLength = map->size - 1;	//use the even length (from the 2^n part) since we can halve it each time
	map->square = 1;					//start on square
	map->range = TERRAIN_HEIGHT;
	map->min = 0;
}

void hmap_free(HeightMap* map) {
	free_a2d((void**)map->heights, map->size);
	
	if( map->displayId != 0 ) {
		hmap_decompile(map);
	}
	
	free(map);
}


/**
 * Take one step in the square-diamond 2.5D Midpoint displacement algorithm.
 * 
 * Square:
 *	x  .  .  .  x
 *	.  .  .  .  .
 *	.  .  *  .  .
 *	.  .  .  .  .
 *	x  .  .  .  x
 * 
 * Diamond:
 *	x  .  *  .  x
 *	.  .  .  .  .
 *	*  .  x  .  *
 *	.  .  .  .  .
 *	x  .  *  .  x
 * 
 */
int hmap_step(HeightMap* map) {
	int len, steps;
	int i, j;
	float avg;
	
	//Loop until smallest divisibility
	if( map->stepLength <= 1 ) 
		return 1;	//already done, so do nothing.
	
	
	//Init local state variables
	len = map->stepLength;	//resume where we left off -- this is the length of one square's side at present
	steps = (map->size - 1) / len;
	
	if( map->square ) {
		//Square
		printf("Doing square, len=%d\n", len);
		
		for(i = 0; i < steps; i++) {
			for(j = 0; j < steps; j++) {
				
				int x1 = i*len, x2 = (i+1)*len;
				int y1 = j*len, y2 = (j+1)*len;
				
				avg = (map->heights[x1][y1] + map->heights[x1][y2] + map->heights[x2][y1] + map->heights[x2][y2]) / 4.0;
				map->heights[x1 + len/2][y1 + len/2] = avg + myrand(map->range, -map->range/2.0f);
			}
		}
	} else {
		//Diamond -- may have only 3 points on an edge
		printf("Doing diamond, len=%d\n", len);
		
		for(i = 0; i < steps + 1; i++) {
			for(j = 0; j < steps + 1; j++) {
				int below;
				for(below = 0; below < 2; below++) {
					//Calculate the value below (i,j) and beside it
					int x1, x2, x3, y1, y2, y3;
					
					if( below ) {
						//(i,j) is (x2,y1)
						//printf("...doing below (%d,%d), steps=%d\n",i,j,steps);
						if( j == steps )
							continue;	//no point below here
						
						x1 = i*len - len/2;
						x2 = i*len;
						x3 = i*len + len/2;
						y1 = j*len;
						y2 = j*len + len/2;
						y3 = (j+1)*len;
					} else {
						//(i,j) is (x1,y2)
						//printf("...doing beside (%d,%d), steps=%d\n",i,j,steps);
						if( i == steps )
							continue;	//no point beside here
						
						x1 = i*len;
						x2 = i*len + len/2;
						x3 = (i+1)*len;
						y1 = j*len - len/2;
						y2 = j*len;
						y3 = j*len + len/2;
					}
					
					if( x1 < 0 ) {
						//On left edge
						avg = (map->heights[x2][y1] + map->heights[x3][y2] + map->heights[x2][y3]) / 3.0;
					} else if( x3 >= map->size ) {
						//On right edge
						avg = (map->heights[x2][y1] + map->heights[x2][y3] + map->heights[x1][y2]) / 3.0;
					} else if( y1 < 0 ) {
						//On top edge
						avg = (map->heights[x3][y2] + map->heights[x2][y3] + map->heights[x1][y2]) / 3.0;
					} else if( y3 >= map->size ) {
						//On bottom edge
						avg = (map->heights[x2][y1] + map->heights[x3][y2] + map->heights[x1][y2]) / 3.0;
					} else {
						//Have all four
						avg = (map->heights[x2][y1] + map->heights[x3][y2] + map->heights[x2][y3] + map->heights[x1][y2]) / 4.0;
					}
					
					map->heights[x2][y2] = avg + myrand(map->range, -map->range/2.0f);
				}
			}
		}
		
		//After diamond, reduce length
		map->stepLength = len/2;
	}
	
	//Update state
	map->range *= map->roughness;
	map->square = !map->square;
	
	//Loop until smallest divisibility
	if( map->stepLength <= 1 ) 
		return 1;
	else
		return 0;
}

void hmap_print(HeightMap *map) {
	int i, j;
	
	printf("=== Map (size=%d, len=%d, sqr=%d) ===\n", map->size, map->stepLength, map->square);
	for(i = 0; i < map->size; i++) {
		for(j = 0; j < map->size; j++) {
			printf("%5.1f ", map->heights[i][j]);
		}
		printf("\n");
	}
	printf("\n");
}