#include <stdlib.h>
#include <stdio.h>
#include "ThermalWeather.h"
#include "Util.h"


/**
 * Allocates the space and calls init.
 */
ThermalState* thermal_create(HeightMap* map) {
	ThermalState *thermal;
	int i, j;
	
	thermal = (ThermalState*)malloc(sizeof(ThermalState));
	thermal->map = map;
	thermal->next = (float**)malloc_a2d(map->size, map->size, sizeof(float));
	thermal->T = 10;
	thermal->C = 0.2;
	
	return thermal;
}



void thermal_free(ThermalState *thermal) {
	free_a2d((void**)thermal->next, thermal->map->size);
	free(thermal);
}


/**
 * Take one step in Muskgrave's Thermal Weathering algorithm.
 * 
 */
int thermal_step(ThermalState *thermal) {
	int i,j;
	float **current, **next;
	int change;
	
	//Local vars to shorten code (and not repeat pointer access)
	current = thermal->map->heights;
	next = thermal->next;
	
	change = 0;
	
	//Need to copy current state into next, so we can add deltas to it incrementally (a cell may be updated by multiple neighbors)
	for(i = 0; i < thermal->map->size; i++) {
		for(j = 0; j < thermal->map->size; j++) {
			next[i][j] = current[i][j];
		}
	}
	
	//Update at each cell
	for(i = 0; i < thermal->map->size; i++) {
		for(j = 0; j < thermal->map->size; j++) {
			int sendN, sendE, sendS, sendW, sendTotal;	//amounts to send in each direction, and the total
			int diff;	//tmp used to store height diff between two cells
			int count;
			
			sendTotal = 0;
			count = 0;
			
			if( j > 0 && (diff = current[i][j] - current[i][j-1]) > 0 && diff > thermal->T ) {
				//Have a north cell with lower altitude, and not at angle of repose
				sendN = diff*thermal->C;
				sendTotal += sendN;
				count++;
			} else
				sendN = 0;
			
			if( i < thermal->map->size - 1 && (diff = current[i][j] - current[i+1][j]) > 0 && diff > thermal->T ) {
				//Have an east cell with lower altitude, and not at angle of repose
				sendE = diff*thermal->C;
				sendTotal += sendE;
				count++;
			} else
				sendE = 0;
			
			if( j < thermal->map->size - 1 && (diff = current[i][j] - current[i][j+1]) > 0 && diff > thermal->T ) {
				//Have a south cell with lower altitude, and not at angle of repose
				sendS = diff*thermal->C;
				sendTotal += sendS;
				count++;
			} else
				sendS = 0;
			
			if( i > 0 && (diff = current[i][j] - current[i-1][j]) > 0 && diff > thermal->T ) {
				//Have a west cell with lower altitude, and not at angle of repose
				sendW = diff*thermal->C;
				sendTotal += sendW;
				count++;
			} else
				sendW = 0;
			
			
			if( sendN != 0 ) {
				next[i][j-1] += sendN/count;
			}
			if( sendE != 0 ) {
				next[i+1][j] += sendE/count;
			}
			if( sendS != 0 ) {
				next[i][j+1] += sendS/count;
			}
			if( sendW != 0 ) {
				next[i-1][j] += sendW/count;
			}
			
			if( sendTotal != 0 ) {
				next[i][j] -= sendTotal/count;	//if not zero, count can't be zero either
				change = 1;						//something changed, so report alg as not done
			}
		}
	}
	
	//Swap current and next (same type and size chunk of memory, just allocated elsewhere, though we switch back if #iter is even)
	float** tmp;
	tmp = thermal->map->heights;
	thermal->map->heights = thermal->next;
	thermal->next = tmp;
	
	return !change;
}

void thermal_print(ThermalState *thermal) {
	int i, j;
	
	printf("=== Thermal State's Map (size=%d) ===\n", thermal->map->size);
	for(i = 0; i < thermal->map->size; i++) {
		for(j = 0; j < thermal->map->size; j++) {
			printf("%5.1f ", thermal->map->heights[i][j]);
		}
		printf("\n");
	}
	printf("\n");
}